Attribute VB_Name = "modWicD2d1"
' Autor: F. Schler (frank@activevb.de)
' Datum: 03/2019

Option Explicit

' ----==== Const ====----
Private Const S_OK As Long = &H0&
Private Const GENERIC_READ As Long = &H80000000
Private Const DXGI_FORMAT_B8G8R8A8_UNORM As Long = &H57&
Private Const D2D1_ALPHA_MODE_PREMULTIPLIED As Long = &H1&
Private Const D2D1_BITMAP_INTERPOLATION_MODE_LINEAR As Long = &H1&
Private Const IID_ID2D1Factory1 As String = "{bb12d362-daee-4b9a-aa1d-14ba401cfa1f}"
Private Const IID_IWICImagingFactory As String = "{ec5ec8a9-c395-4314-9c77-54d7a935ff70}"
Private Const CLSID_WICImagingFactory2 As String = "{317d06e8-5f24-433d-bdf7-79ce68d8abc2}"
Private Const GUID_WICPixelFormat32bppPBGRA As String = "{6fddc324-4e03-4bfe-b185-3d77768dc910}"

' ----==== Enums ====----
Private Enum vtb_Interfaces
    CreateDecoderFromFilename = 3
    GetSize = 4
    CreateBitmapFromWicBitmap = 5
    Initialize = 8
    CreateFormatConverter = 10
    GetFrame = 13
    CreateHwndRenderTarget = 14
    DrawBitmap = 26
    Clear = 47
    BeginDraw = 48
    EndDraw = 49
    Resize = 58
End Enum

' ----==== Types ====----
Private Type GUID
    data1 As Long
    data2 As Integer
    data3 As Integer
    data4(0 To 7) As Byte
End Type

Private Type RGBA
    r As Byte
    g As Byte
    b As Byte
    a As Byte
End Type

Private Type D2D1_PIXEL_FORMAT
    dFormat As Long
    alphaMode As Long
End Type

Private Type D2D1_SIZE_U
    width As Long
    height As Long
End Type

Private Type D2D1_SIZE_F
    width As Single
    height As Single
End Type

Private Type D2D1_RECT_F
    left As Single
    top As Single
    right As Single
    bottom As Single
End Type

Private Type D2D1_COLOR_F
    r As Single
    g As Single
    b As Single
    a As Single
End Type

Private Type D2D1_RENDER_TARGET_PROPERTIES
    eType As Long
    pixelFormat As D2D1_PIXEL_FORMAT
    dpiX As Single
    dpiY As Single
    usage As Long
    minLevel As Long
End Type

Private Type D2D1_HWND_RENDER_TARGET_PROPERTIES
    hwnd As Long
    pixelSize As D2D1_SIZE_U
    presentOptions As Long
End Type

' ----==== Ole32 API-Deklarationen ====----
Private Declare Function CLSIDFromString Lib "ole32.dll" ( _
                         ByVal pstring As Long, _
                         ByRef pCLSID As GUID) As Long

' ----==== Kernel32.dll API-Deklarationen ====----
Private Declare Sub CopyMemory Lib "kernel32.dll" _
                    Alias "RtlMoveMemory" ( _
                    ByRef Destination As Any, _
                    ByRef Source As Any, _
                    ByVal length As Long)

' ----==== D2d1.dll API-Deklarationen ====----
Private Declare Function D2D1CreateFactory Lib "D2d1.dll" ( _
                         ByVal factoryType As Long, _
                         ByVal riid As Long, _
                         ByVal pFactoryOptions As Long, _
                         ByRef ppIFactory As Long) As Long

Public Function CreateFromFilename(ByVal sFileName As String, ByVal frm As Form) As Boolean

    Dim bRet As Boolean

    Dim psFileName As Long
    Dim psID2D1Factory1 As Long
    Dim psWICPixelFormat As Long

    Dim pBitmap As Long
    Dim pBmpFrame As Long
    Dim pBmpDecoder As Long
    Dim pWicFactory As Long
    Dim pD2D1Factory As Long
    Dim pBmpConverter As Long
    Dim pRenderTarget As Long

    Dim ppBitmap As Long
    Dim ppBmpFrame As Long
    Dim ppBmpDecoder As Long
    Dim ppBmpConverter As Long
    Dim ppRenderTarget As Long

    Dim tGUID As GUID
    Dim tSIZEF As D2D1_SIZE_F
    Dim tSIZEU As D2D1_SIZE_U
    Dim tRECTF As D2D1_RECT_F
    Dim tCOLORF As D2D1_COLOR_F
    Dim tRTP As D2D1_RENDER_TARGET_PROPERTIES
    Dim tHRTP As D2D1_HWND_RENDER_TARGET_PROPERTIES

    Dim ptRTP As Long
    Dim ptHRTP As Long
    Dim ptGUID As Long
    Dim ptSIZEF As Long
    Dim ptSIZEU As Long
    Dim ptRECTF As Long
    Dim ptCOLORF As Long

    bRet = False
    
    frm.ScaleMode = vbPixels
    frm.BackColor = &HF0F0F0

    psID2D1Factory1 = StrPtr(IID_ID2D1Factory1)

    If CLSIDFromString(psID2D1Factory1, tGUID) = S_OK Then

        ptGUID = VarPtr(tGUID)

        ' D2D1 Factory erstellen
        If D2D1CreateFactory(0&, ptGUID, 0&, pD2D1Factory) = S_OK Then

            ' Pixelformat fr das D2D1 RenderTarget festlegen
            With tRTP
                
                .pixelFormat.dFormat = DXGI_FORMAT_B8G8R8A8_UNORM
                .pixelFormat.alphaMode = D2D1_ALPHA_MODE_PREMULTIPLIED

            End With

            ' Hwnd, auf dem gerendert werden soll, festlegen
            With tHRTP

                .hwnd = frm.hwnd

            End With

            ptRTP = VarPtr(tRTP)
            ptHRTP = VarPtr(tHRTP)
            ppRenderTarget = VarPtr(pRenderTarget)

            ' D2D1 RenderTarget vom Hwnd erstellen
            If Invoke(pD2D1Factory, CreateHwndRenderTarget, ptRTP, ptHRTP, ppRenderTarget) = S_OK Then

                ' Gre des RenderTargets festlegen
                With tSIZEU

                    .width = frm.ScaleWidth
                    .height = frm.ScaleHeight

                End With

                ptSIZEU = VarPtr(tSIZEU)

                ' D2D1 RenderTargets Resize
                If Invoke(pRenderTarget, Resize, ptSIZEU) = S_OK Then

                    ' WIC Factory erstellen
                    If InitInterface(pWicFactory, CLSID_WICImagingFactory2, IID_IWICImagingFactory) = True Then

                        psFileName = StrPtr(sFileName)
                        ppBmpDecoder = VarPtr(pBmpDecoder)

                        ' WIC Decoder von der Datei erstellen
                        If Invoke(pWicFactory, CreateDecoderFromFilename, psFileName, 0&, GENERIC_READ, 0&, ppBmpDecoder) = S_OK Then

                            ppBmpFrame = VarPtr(pBmpFrame)

                            ' WIC FrameDecode erstellen (erstes Frame)
                            If Invoke(pBmpDecoder, GetFrame, 0&, ppBmpFrame) = S_OK Then

                                ppBmpConverter = VarPtr(pBmpConverter)

                                ' WIC FormatConverter erstellen
                                If Invoke(pWicFactory, CreateFormatConverter, ppBmpConverter) = S_OK Then

                                    psWICPixelFormat = StrPtr(GUID_WICPixelFormat32bppPBGRA)

                                    If CLSIDFromString(psWICPixelFormat, tGUID) = S_OK Then

                                        ptGUID = VarPtr(tGUID)
                                      
                                        ' WIC Bitmap in das 32bppPBGRA Pixelformat konvertieren
                                        If Invoke(pBmpConverter, Initialize, pBmpFrame, ptGUID, 0&, 0&, CDbl(0), 0&) = S_OK Then

                                            ppBitmap = VarPtr(pBitmap)
                                            
                                            ' ein D2D1 Bitmap vom WIC Bitmap erstellen
                                            If Invoke(pRenderTarget, CreateBitmapFromWicBitmap, pBmpConverter, 0&, ppBitmap) = S_OK Then
                                                
                                                ' Mit dem zeichnen auf dem D2D1 RenderTaget beginnen
                                                Call Invoke(pRenderTarget, BeginDraw)

                                                ' Hintergrundfarbe festlegen
                                                tCOLORF = COLOR_F_LONGA(frm.BackColor, 1)

                                                ptCOLORF = VarPtr(tCOLORF)

                                                ' lscht den Inhalt vom D2D1 RenderTarget mit einer Hintergrundfarbe
                                                Call Invoke(pRenderTarget, Clear, ptCOLORF)

                                                ptSIZEF = VarPtr(tSIZEF)

                                                ' Gre des D2D1 Bitmaps ermitteln
                                                Call Invoke(pBitmap, GetSize, ptSIZEF)

                                                ' Position und Gre zum zeichnen festlegen
                                                With tRECTF

                                                    .left = 20
                                                    .top = 20
                                                    .right = .left + tSIZEF.width
                                                    .bottom = .top + tSIZEF.height

                                                End With

                                                ptRECTF = VarPtr(tRECTF)
                                                
                                                ' zeichnet das D2D1 Bitmap in den D2D1 RenderTarget an entsprechnder Position
                                                If Invoke(pRenderTarget, DrawBitmap, pBitmap, ptRECTF, CSng(1), _
                                                    D2D1_BITMAP_INTERPOLATION_MODE_LINEAR, 0&) = S_OK Then

                                                    bRet = True

                                                End If

                                                ' zeichnen in das D2D1 RenderTarget beenden
                                                Call Invoke(pRenderTarget, EndDraw, 0&, 0&)
                                                
                                                ' Aufrumen
                                                Call modInterface.ReleaseInterface(pBitmap)

                                            End If

                                        End If

                                    End If

                                End If

                                ' Aufrumen
                                Call modInterface.ReleaseInterface(pBmpConverter)

                            End If

                            ' Aufrumen
                            Call modInterface.ReleaseInterface(pBmpFrame)

                        End If

                        ' Aufrumen
                        Call modInterface.ReleaseInterface(pBmpDecoder)

                    End If

                    ' Aufrumen
                    Call modInterface.ReleaseInterface(pWicFactory)

                End If

                ' Aufrumen
                Call modInterface.ReleaseInterface(pRenderTarget)

            End If

            ' Aufrumen
            Call modInterface.ReleaseInterface(pD2D1Factory)

        End If

    End If

    CreateFromFilename = bRet

End Function

Private Function Invoke(ByVal pInterface As Long, ByVal vtb As vtb_Interfaces, _
    ParamArray var()) As Variant

    If pInterface <> 0& Then

        Invoke = modInterface.OleInvoke(pInterface, vtb, var)

    End If

End Function

Private Function COLOR_F_LONGA(ByVal lColor As Long, Optional ByVal a As Single = 1!) As D2D1_COLOR_F

    Dim tRGBA As RGBA

    Call CopyMemory(tRGBA, lColor, 4)

    With COLOR_F_LONGA

        .a = a
        .r = CSng(tRGBA.r) / 255!
        .g = CSng(tRGBA.g) / 255!
        .b = CSng(tRGBA.b) / 255!

    End With

End Function

